/*
 * Implementation for the buffered read function.
 */
	

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */
 
#include "cleansocks.h"
#include "cleanbuf.h"

#include <string.h>

namespace cleansocks {

#ifdef WIN32
static int initted = sockinit();
#endif

/*
 * Move from the buffer to ubuf, which is size ubufsize, up to term.  (If
 * term < 0 there is no terminator.)  If peek is true, copies instead of
 * moving.  Returns a structure that tells the number of bytes transferred,
 * and a boolean telling if the terminator was found.
 */
buffered_socket::unbufret_t buffered_socket::unbuffer
		(char *ubuf, int ubufsize, int term, bool peek /* = false */)
{
	int ret = 0;
	char *nl = NULL;

	/* See if there is buffered material to transfer into the
	   receive area. */
	if(m_validsize > 0) {
		/* Find the right amount of the existing buffer to return: All
		   of it, or up to the receive area limit, or to the first
		   term (usually newline). */
		ret = m_validsize;
		if(ret > ubufsize) ret = ubufsize;
		nl = term >= 0 ? 
			static_cast<char *>(memchr(m_valid, term, ret)) : NULL;
		if(nl) ret = (nl - m_valid) + 1;

		/* Transfer that amount. */
		memcpy(ubuf, m_valid, ret);
		if(!peek) {
			m_validsize -= ret;
			m_valid += ret;
		}
	}
	
	// Return number of characters, and if terminator found.
	return unbufret_t(ret, nl != NULL);
}


int buffered_socket::recvln(void *_ubuf, int ubufsize, int term, int flags)
{
	char *ubuf(static_cast<char *>(_ubuf));

	// Record peekiness.
	bool peek = flags & MSG_PEEK;

	// Try to satisfy from the buffer.
	unbufret_t ubr = unbuffer(ubuf, ubufsize, term, peek);
	if(ubr.num > 0 && (ubr.num >= ubufsize || ubr.term || m_err_pend < 1))
		return ubr.num;

	// If there is a read error pending, time to deliver it.  (There must
	// have been no buffered bytes.)
	if(m_err_pend < 1) {
		if(m_err_pend == 0) return 0;
		throw m_pending;
	}

	// Account for the characters added to the return buffer.
	int ret = ubr.num;
	ubuf += ret;
	ubufsize -= ret;

	// We don't want to send this to the system, since we might then 
	// wait for too much data.  We'll just keep issuing reads until
	// we fill the user buffer (unless we see a term).
	bool waitall = flags & MSG_WAITALL;
	flags &= ~MSG_WAITALL;
	
	// Note: The buffer must now be empty.  For control to reach here,
	// either ubr.num == 0, in which case the buffer must have started
	// empty, otherwise, ubr.num < ubufsize && !ubr.term must be true.
	// In that case, there was room for all the buffered data, and no
	// terminator was found.

	// If we got some bytes from the buffer, and waitall is false,
	// and we're not looking for a terminator (which we haven't found
	// or we would have returned already), we need to make sure the actual
	// read won't block, since there might not be any more data, so we
	// would block without returning what we have.
	if(ret > 0 && !waitall && term == -1) {
#ifdef WIN32
		// As usual, Winsock is broken: no MSG_DONTWAIT.  There 
		// being no requirement to fill the buffer, we'll just 
		// return rather than risk a hang.
		return ret;
#else
		// If we're unable to send flags to the lower layer,
		// we won't be able to send DONTWAIT.  That's true
		// if we're using the horrid capability interface
		// adaptor object.  If so, we have to be as broken as
		// Windows, even on a real OS.
		if(m_srf2sr.valid()) return ret;

        	// Linux being a functional OS, we'll just make sure the 
        	// read doesn't wait, but get any bytes that happen to
		// have arrived and fit in the buffer.
		flags |= MSG_DONTWAIT;
#endif // WIN32
	}

	// Okay.  Let's try actually reading some data.
	while(true) {
		// Find where and how much to read.  We'll read into the larger
		// of the user area and the buffer, except that if we're
		// looking for a terminator we can't use the user buffer
		// because we might have extra to keep.  If we're peeking,
		// we'll always read to the user buffer because we never want
		// to keep it.
		char *target = m_buf;
		int readsize = m_allosize;
		bool bufread = true;
		if(peek || (term < 0 && ubufsize > readsize)) {
			// Read to the user buffer.  
			target = ubuf;
			readsize = ubufsize;
			bufread = false;
		}

		// Read the connection.  
		int nread = 0;
		try {
			nread = cleansocks::recv(*m_socket, target, 
						 readsize, clean_flags(flags));
		} catch (socket_would_block &b) {
			// Must be using non-blocking mode.  If we got something
			// from the buffer, return that.  Otherwise, forward
			// the exception.
			if(ret > 0) return ret;
			throw b;
		} catch(socket_sys_error &e) {
			// Similar, but we forward now if there's no data
			// already, otherwise save to later.
			if(ret == 0) throw e;
			m_err_pend = -1;
			m_pending = e;
		}
		if(nread == 0) {
			// EOF.  If we're returning previous bytes, save for
			// later and return them.  Else, return the EOF now.
			if(ret > 0) m_err_pend = 0;
			return ret;
		}

		// Which way do we move?
		if(bufread) {
			// Move stuff out of the holding buffer to the return
			// area.
			m_valid = m_buf;
			m_validsize = nread;
			unbufret_t ubr = unbuffer(ubuf, ubufsize, term);
			if(ubr.term) return ret + ubr.num;
			nread = ubr.num;
		} else {
			// If there is a terminator, see if it was found, and,
			// if so, trim the request to the data up to the
			// terminator.
			if(term >= 0) {
				char *nl = static_cast<char *>
					(memchr(target, term, nread));
				if(nl) {
					// Extra chars in user buffer, past
					// terminator.  We need to reduce the
					// return size, and in normal mode
					// (not peeking), we copy extras into
					// the buffer for later reuse.
					int keep = (nl - target) + 1;
					if(!peek) {
						m_validsize = nread - keep;
						m_valid = m_buf;
						memcpy(m_buf, target + keep,
						       m_validsize);
					}
					return ret + keep;
				}
			}
		}

		// Adjust the return count, return buffer pointer, and
		// return buffer size to account for the number of bytes
		// added to it by the read.
		ret += nread;
		ubuf += nread;
		ubufsize -= nread;

		// We return if we've filled the user buffer, if there's an
		// error pending (we don't want to read again), or we don't
		// need to keep reading because we're not trying to fill the
		// buffer and we're not looking for a terminator character.
		if(ubufsize <= 0 || m_err_pend < 1 || (!waitall && term < 0))
			return ret;

		// Go 'round again.
	}
}

}

