/*
 * Implementations for the IP4 version of Clean Socks.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#include "cleansocks.h"
#include "cleanip.h"

#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#endif

#include <string>

namespace cleansocks {

#ifdef WIN32
static int initted = sockinit();
#endif

socket::sock_descr_t TCPsocket::sm_descr(AF_INET, SOCK_STREAM, "tcp");
socket::sock_descr_t UDPsocket::sm_descr(AF_INET, SOCK_DGRAM, "udp");

IPaddress lookup_host(const char *hostname)
{
	// Perform the lookup.
	struct addrinfo hints, *result;
	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_INET;
	int rtn = getaddrinfo(hostname, NULL, &hints, &result);

	// Check for failures.
	socket_db_error::check_rtn
		("IPaddress::lookup("+std::string(hostname)+")", rtn);
	if(result == NULL) throw socket_error("Null name info");
	
	// Extract, clean up, and return.
	IPaddress ret;
	ret.m_addr = reinterpret_cast<sockaddr_in *>
		(result->ai_addr)->sin_addr.s_addr;
	freeaddrinfo(result);
	return ret;
}

inline int againfilt(int ret)
{
	if(ret < 0) {
		int e = ERRCODE;
#ifdef WIN32
		if(e == WSAEWOULDBLOCK)
#else
		if(e == EAGAIN || e == EWOULDBLOCK)
#endif
			throw socket_would_block("send/recv",e);
		throw socket_sys_error("send/recv",e);
	}

	return ret;
}

int TCPsocket::send(const void *buffer, int size, int flags)
{
	int ret = ::send(sock(), V2C(buffer), size, flags | MSG_NOSIGNAL);
	return againfilt(ret);
}
int TCPsocket::recv(void *buffer, int size, int flags)
{
	int ret = ::recv(sock(), V2C(buffer), size, flags);
	return againfilt(ret);
}
int UDPsocket::sendto(const void *buffer, int size, int flags,
		      const endpoint &dest)
{
	int ret = ::sendto(sock(), V2C(buffer), size, flags,
			   dest.addr(), dest.addrlen());
	return againfilt(ret);
}
int UDPsocket::recvfrom(void *buffer, int size, int flags, endpoint *from)
{
	int ret = ::recvfrom(sock(), V2C(buffer), size, flags,
			     from ? from->addr() : NULL,
			     from ? &from->addrlen() : NULL);
	return againfilt(ret);
}

// IP address to dotted dec string.
std::string IPaddress::str() const
{
	uint32_t ip = ntohl(m_addr);
	int shift = 24;
	std::string ret;
	for(int i = 0; i < 4; ++i) {
		ret += std::to_string(ip >> shift & 0x0ff);
		shift -= 8;
		if(i < 3) ret += ".";
	}

	return ret;
}

IPport lookup_service(const char *service)
{
	// Perform the lookup.
	struct addrinfo *result;
	int rtn = getaddrinfo(NULL, service, NULL, &result);

	// Check for errors.
	if(result == NULL) throw socket_error("Null name info");
	socket_db_error::check_rtn("IPport::lookup("+std::string(service)+")", rtn);

	// Collect, clean up and return.
	IPport ret;
	ret.m_port = reinterpret_cast<sockaddr_in *>(result->ai_addr)->sin_port;
	freeaddrinfo(result);
	return ret;
}

IPendpoint::IPendpoint()
{
	m_saptr = reinterpret_cast<struct sockaddr *>(&m_sa);
	m_size = sizeof m_sa;
	memset(&m_sa, 0, sizeof m_sa);
}

IPendpoint::IPendpoint(const IPaddress &addr, const IPport &port)
{
	m_saptr = reinterpret_cast<struct sockaddr *>(&m_sa);

	memset(&m_sa, 0, sizeof m_sa);
	m_sa.sin_family = AF_INET;
	m_sa.sin_port = port.m_port;
	m_sa.sin_addr.s_addr = addr.m_addr;
	m_size = sizeof(struct sockaddr_in);
}

}

#ifdef WIN32
// AFAK, this is some sort of error in the cross-compile.  This function is
// not present in the mingw libraries, and not implemented in the relevant
// headers for the cross-compile from Fedora.  This:
//   http://mingw.5.n7.nabble.com/Re-Problem-gai-strerrorA-undefined-reference-td32775.html
// suggests it's supposed to be in-line.  The following code I found here:
//   https://searchcode.com/codesearch/view/20369985/
// I says it belongs to pkgs.fedoraproject.org/mingw-crt
char *gai_strerrorA(int ecode)
{
	DWORD dwMsgLen __attribute__((unused));
	static char buff[GAI_STRERROR_BUFFER_SIZE + 1];
	dwMsgLen = FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_MAX_WIDTH_MASK,
				  NULL, ecode, MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT), (LPSTR)buff,
				  GAI_STRERROR_BUFFER_SIZE, NULL);
	return buff;
}

// This is on the order of "should probably compile and not crash when
// it runs."  I'll take that.  All quite strange.
#include <wchar.h>
#if 0
wchar_t *gai_strerrorW(int ecode)
{
	char *ret = gai_strerrorA(ecode);
	static wchar_t buff[GAI_STRERROR_BUFFER_SIZE + 1];
	int i;
	for(i = 0; i < GAI_STRERROR_BUFFER_SIZE; ++i) {
		buff[i] = ret[i];
		if(!ret[i]) break;
	}
	return buff;
}
#endif

// From
//https://searchcode.com/file/14152807/%20railoapacheportable/apache/include/tpipv6.
// Might be rather better, if it works.
WCHAR *gai_strerrorW(int ecode)
{
    DWORD dwMsgLen;
    static WCHAR buff[GAI_STRERROR_BUFFER_SIZE + 1];

    dwMsgLen = FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM
                             |FORMAT_MESSAGE_IGNORE_INSERTS,
                              NULL,
                              ecode,
                              MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                              (LPWSTR)buff,
                              GAI_STRERROR_BUFFER_SIZE,
                              NULL);

    return buff;
}



#endif
