/*
 * This contains the small number of implementations for the main Clean
 * Sockets library.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#include "cleansocks.h"

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#endif

#include <string>

namespace cleansocks {

#ifdef WIN32

// Winsock requires running an initialization function.  This runs it, with a
// flag that allows it to be safely re-run without effect.  The library 
// arranges to make sure init() is called using static initializations like
// the one below here, one in each .cpp file file, to make sure it is called
// before use.  A race is still possible that would make multiple calls of
// WSAStartup in the same process, but this is apparenlty allowed.
int sockinit()
{
	static bool initted = false;
	if(initted) return 1;

	static WSADATA wsaData;
	if(WSAStartup(MAKEWORD(2,2), &wsaData))
		throw socket_error("Winsocks annoying initialization failed.");
	initted = true;
	return 1;
}
static int initted = sockinit();
#endif

// Look up the protocol number from its name.
// The Windows docs say that getprotobyname allocates one copy of the protoent
// structure per thread, so it should be thread safe.  The Unix version is not,
// and there seems not to be a "standard" replacement.  The one used here is
// from the GNU toolchain.  Theoretically, it could fail because the buffer is
// not big enough, and the documentation says to try again with a bigger one.
// Poo.
int socket::lookup_proto(const char *name)
{
	struct protoent *pe = NULL;
#if defined(WIN32) || defined(BSD) || defined(NOGETPROTOBYNAME_R)
	pe = getprotobyname(name);
#else
	struct protoent protoent;
	char buffer[2048];
	getprotobyname_r(name, &protoent, buffer, sizeof buffer, &pe);
#endif
	if(pe == NULL)
		throw socket_error(std::string("Protocol ") +
				   name + " not found");
	return pe->p_proto;
}
#ifdef WIN32
// Winsock supports an odd set set of getsockopt keys.  Someone in
// Redmond must sit up nights finding peculiar ways to make simple
// things complicated.
// http://msdn.microsoft.com/en-us/library/windows/desktop/ms740532%28v=vs.85%29.aspx
socket::sock_descr_t socket::get_sock_descr() const
{
	if(m_sock == INVALID_SOCKET) 
		throw socket_error("Socket not typed");

	WSAPROTOCOL_INFO proti;
	int pilen = sizeof proti;
	int ret = ::getsockopt(m_sock, SOL_SOCKET, SO_PROTOCOL_INFO,
			       (char*)&proti, &pilen);
	if(ret != 0) {
		int e = ERRCODE;
		throw socket_sys_error("getsockopt",e);
	}
	return sock_descr_t(proti.iAddressFamily, proti.iSocketType,
			    proti.iProtocol);
}
#else
// Code for functional systems.
inline int socket::getopt(int opt) const
{
	int ret;
	socklen_t size = sizeof ret;
	if(::getsockopt(m_sock, SOL_SOCKET, opt, &ret, &size) < 0) {
		int e = ERRCODE;
		throw socket_sys_error("getsockopt",e);
	}
	return ret;
}
#ifdef BSD
socket::sock_descr_t socket::get_sock_descr() const
{
	if(m_sock == INVALID_SOCKET) 
		throw socket_error("Socket not typed");
	check_broken();
	return sock_descr_t(m_domain, getopt(SO_TYPE), m_protocol);
}
#else	
socket::sock_descr_t socket::get_sock_descr() const
{
	if(m_sock == INVALID_SOCKET) 
		throw socket_error("Socket not typed");
	return sock_descr_t(getopt(SO_DOMAIN), getopt(SO_TYPE),
			    getopt(SO_PROTOCOL));
}
#endif // not BSD	
#endif // not WIN32

void socket::assert_descr(const sock_descr_t &descr) const
{
	sock_descr_t me = get_sock_descr();

	if(me.domain == descr.domain && me.type == descr.type &&
	   me.protocol == descr.protocol)
		return;

	throw socket_interface_error("Illegal socket cast");
}

#ifdef WIN32
// Consulted:
// http://bobobobo.wordpress.com/2009/02/02/getting-winsock-error-messages-in-string-format/
// Whole lot more useful than the official stuff from Microsoft.
std::string socket_sys_error::set_code(int code)
{
	// Error code.  Also record in the object
	int errMsg = code;
	m_code = errMsg;
	
	LPSTR errString = NULL;
	
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_ALLOCATE_BUFFER, 
		0, 			// Null source
		errMsg,		// Code to interpret
		0, 			// Auto-determine language.
		(LPSTR)&errString, // Put it here. 
		0, NULL);	// Buf size and extra fill-in arguments (none).
		
	std::string retval(errString);
	LocalFree(errString);
	return retval;
}
#endif
}
