#!/usr/bin/perl
# This extracts the error codes from the file on the command line (which is expected to be the
# openssl source file x509_vfy.h).  It generates C code to declare an array of plain C
# strings to give the name for each number.  Exits zero if it thinks it succeeded.  Even if
# it fails, it generates a legal definition for an empty array.
use strict;

sub genempty() {
    print "static const char *cert_err_list[1] = { NULL }; \n";
    print "static const int cert_err_last = -1;\n";

    exit 1;
}

my $fn = shift @ARGV;
my $fh;
if(!open($fh, $fn)) { genempty(); }

# I'm going to read the codes.  I've decided not to assume they start at 1 or are listed
# consecutively (as the are), but just that they are pretty dense and don't start very large.
# Thought that might be less brittle.

my %codes = { };
my $max = -1;
my $count = 0;

while(my $ln = <$fh>) {
    chomp $ln;

    if($ln =~ m@^\s*\#\s*define\s+X509_V_ERR_([A-Z0-9_]+)\s+(\d+)\s*(/\*.*\*/\s*)?$@) {
	my ($name, $num) = ($1, $2);
	$codes{$num} = $name;
	if($max < $num) { $max = $num; }
	++$count;
    }
}

# Sanity limits.
if($count == 0 || $max > 200) { genempty(); }

print "static const int cert_err_last = ", $max, ";\n";
print "static const char *cert_err_list[", $max+1, "] = {\n";
for(my $i = 0; $i <= $max; ++$i) {
    if(exists $codes{$i}) {
	print '  "', $codes{$i}, '"', ",\n";
    } else {
	print "  NULL,\n";
    }
}
print "};\n";

exit 0;

