/*
 * Run with a single HTTP URL on the command line.  Will attempt to fetch it
 * as HTTP/1.0 and print the results.  URL parsing is none-too-sophisticated,
 * so don't expect much.  
 *
 *   buftest1 [ -p ] url [ read-buffer-size ] [ header-buffer-size ] [ norand ]
 *
 * The size of the buffer in the buffered_socket, the size used to read the
 * headers (ditto).  Since those are line-oriented, a small value breaks up
 * header lines.  Normally, the body read size is randomized, with the buffer
 * size as max.  If the third arg is norand, this ranomization is suppressed.
 *
 * The -p flag invokes a test of the peek flag on the reads.  Each read is
 * first performed with a peek, and then compared to the real read.
 *
 * Normally, the program accepts a http URL, but compiled with -DTLS,
 * accepts only https.
 */
	
/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#include <stdlib.h>
	
#include <iostream>
#include <iomanip>
#include <string>
#include <algorithm>
#include <stdexcept>

#include "cleansocks.h"
#include "cleanip.h"
#include "cleanbuf.h"

#ifdef TLS
#include "cleantlsc.h"
#endif

using namespace std;
using namespace cleansocks;

/* Here are a wappers for recv() recvln() that performs the peek test if
   required.  Otherwise, they just run the underlying and return. */
static int peeksize = -1;	// Size of buffer.
static char *peekbuf = NULL;	// Buffer.
static int peekvalid = -1;	// Valid portion of buffer.
static int peekcount = 0;	// Count of tests.
void keep(void *data, int size)
{
	if(peeksize < size) {
		delete [] peekbuf;
		peekbuf = new char [size];
		peeksize = size;
	}
	std::copy((char *)data, (char *)data + size, peekbuf);
	peekvalid = size;
	++peekcount;
}
void check(void *data, int size)
{
	if(size < peekvalid ||
	   !std::equal((char *)data, (char *)data + peekvalid, peekbuf)) {
		std::cout << "**** Peek Test " << peekcount
			  << " Failed ****" << std::endl;
		std::cout << "Pk: " << string(peekbuf,peekvalid) << std::endl;
		std::cout << "Rd: " << string((char*)data,size) << std::endl;
	}
}
int dorecv(srf_obj &sro, void *buf, int size, bool peektest)
{
	if(!peektest)
		return recv(sro, buf, size);
	
	int amt = recv(sro, buf, size, CSMSG_PEEK);
	keep(buf, amt);
	amt = recv(sro, buf, size);
	check(buf, amt);
	return amt;
}
int dorecvln(srfln_obj &sro, void *buf, int size, bool peektest)
{
	if(!peektest)
		return recvln(sro, buf, size);

	int amt = recvln(sro, buf, size, CSMSG_PEEK);
	keep(buf, amt);
	amt = recvln(sro, buf, size);
	check(buf, amt);
	return amt;
}

void doit(string url, int bufsize, int hdrsize, bool dorand, bool peektest)
{
	// Check and remove the leading protocol and slashes, allowing only http or
	// https per compile flag.
#ifdef TLS
	const string prefix = "https://";
#else
	const string prefix = "http://";
#endif
	const int len = prefix.length();
	if(url.length() < len || url.substr(0,len) != prefix)
		throw runtime_error(url + " does not start with \""+prefix+"\"");
	url.erase(0,len);

	// Default port
#ifdef TLS	
	int port = 443;
#else
	int port = 80;
#endif	

	// Hostname, equal to the (stripped) url until we find out otherwise.
	string hostname = url;

	// Isolate the host name, if it's not the entire URL.
	int hend = url.find_first_of("/:");
	if(hend != string::npos) {
		hostname = url.substr(0,hend);

		// Extract a port if present.
		if(url[hend] == ':') {
			port = stoi(url.substr(hend+1));
			hend = url.find("/", hend+1);
		}
	}

	// The rest of the URL is the path, or it is /.
	string path = (hend == string::npos ? "/" : url.substr(hend));

	cout << "host = " << hostname << ", port = " << port
	     << ", path = " << path << endl;

	// Look up the host and try to connect.
	IPaddress a = lookup_host(hostname);
	IPport p = port;
#ifdef TLS
	TCPsocket ts;
	connect(ts,IPendpoint(a,p));
	client_tls_socket s(ts,hostname);
#else	
	TCPsocket s;
	connect(s,IPendpoint(a,p));
#endif
	
	// Make a buffered socket for doing the I/O.
	buffered_socket bsock(s, bufsize);
	
	// Write an HTTP request to the server.
	send(bsock, "GET "+path+" HTTP/1.0\r\n");
	send(bsock, "Host: "+hostname+"\r\n\r\n");

	// Read the header a line at a time.
	char buf[500];
	int n;
	cout << "=== Headers ===" << endl;
	int ct = 0;

	bool suppress = false;

	if(hdrsize <= 0 || hdrsize > sizeof buf) hdrsize = sizeof buf;
	while((n = dorecvln(bsock, buf, hdrsize-1, peektest)) > 0) {
		buf[n] = 0;
		bool done = (buf[n-1] == '\n');

		if(buf[n-2] == '\r') buf[n-2] = 0;
		if(buf[n-1] == '\r') buf[n-1] = 0;

		cout << "[" << setw(2) << n << "] " << buf << endl;

		if(done && !suppress && n <= 2) break;
		suppress = !done;

		++ct;

	}
	cout << "=== End Headers (" << ct << " reads) ===" << endl;

	if(n == 0) {
		cout << "=== End Transmission ===" << endl;
	} else {
		cout << "=== Start Body ===" << endl;

		// Read and print the body.  Use random buffer sizes.
		srand(bufsize);
		ct = 0;
		int reduction = dorand ? rand() % ((sizeof buf) - 10) : 0;
		while((n = dorecv(bsock, buf, (sizeof buf) - reduction,
				  peektest)) != 0) {
			cout.write(buf, n);
			++ct;
		}
		cout << "\n=== End Body (" << ct << " reads)===" << endl;
	}
	close(s);
}

int main(int argc, char **argv)
{
	if(argc < 2 || argc > 6) {
		cerr << argv[0] << " [ -p ] url [ bufsize ] [ hdrsize ] "
		     << "[ norand ]" << endl;
		exit(2);
	}

	try {
		bool peektest = (argc > 1 && string(argv[1]) == "-p");
		if(peektest) {
			++argv;
			--argc;
		}
		doit(argv[1],
		     argc >= 3 ? atoi(argv[2]) : buffered_socket::DEF_BUF_SIZE,
		     argc >= 4 ? atoi(argv[3]) : 0,
		     argc >= 5 ? string(argv[4]) != "norand" : true,
		     peektest);
	} catch(socket_db_error e) {
		cout << "Socket db error: " << e.what() << endl;
	}
	catch(socket_error e) {
		cout << "Socket error: " << e.what() << endl;
	}
	catch(runtime_error e) {
		cout << "Runtime error: " << e.what() << endl;
	}
}

